# The next line is a comment in Tcl, but not sh \
        exec wish $0 -f ${1+"$@"}


puts -nonewline stdout {1}
flush stdout

set w .
set can .can
set status .f.status
set pi 3.1415926
set ufillno 12
set psfileno 1

wm title $w "BDwish"

set palette {black blue green cyan red magenta white yellow palegreen orange dodgerblue gray lightcyan lightgray peru violet}

# for gray scale. Needs more work though.
# set pscolors { \
# {0.000 0.000 0.000 setrgbcolor} \
# {0.850 0.850 0.850 setrgbcolor} \
# {0.000 1.000 0.000 setrgbcolor} \
# {1.000 1.000 0.000 setrgbcolor} \
# {0.300 0.300 0.300 setrgbcolor} \
# {0.750 0.000 1.000 setrgbcolor} \
# {1.000 1.000 1.000 setrgbcolor} \
# {1.000 1.000 0.000 setrgbcolor} \
# {0.500 1.000 0.500 setrgbcolor} \
# {1.000 0.500 0.000 setrgbcolor} \
# {0.000 0.500 1.000 setrgbcolor} \
# {0.750 0.750 0.750 setrgbcolor} \
# {0.750 1.000 1.000 setrgbcolor} \
# {0.800 0.800 0.800 setrgbcolor} \
# {0.750 0.500 0.250 setrgbcolor} \
# {1.000 0.750 1.000 setrgbcolor} }

 set pscolors { \
 {0.000 0.000 0.000 setrgbcolor} \
 {0.000 0.000 1.000 setrgbcolor} \
 {0.000 1.000 0.000 setrgbcolor} \
 {0.000 1.000 1.000 setrgbcolor} \
 {1.000 0.000 0.000 setrgbcolor} \
 {1.000 0.000 1.000 setrgbcolor} \
 {1.000 1.000 1.000 setrgbcolor} \
 {1.000 1.000 0.000 setrgbcolor} \
 {0.500 1.000 0.500 setrgbcolor} \
 {1.000 0.500 0.000 setrgbcolor} \
 {0.000 0.500 1.000 setrgbcolor} \
 {0.750 0.750 0.750 setrgbcolor} \
 {0.750 1.000 1.000 setrgbcolor} \
 {0.800 0.800 0.800 setrgbcolor} \
 {0.750 0.500 0.250 setrgbcolor} \
 {1.000 0.750 1.000 setrgbcolor} }

set i 0
foreach name $palette {
  set pspalette($name) [lindex $pscolors $i]
  incr i 1
}

proc docommand {} {
#  removeinput stdin
  if {[gets stdin command]==-1} {exit}
  if {[info procs [lindex $command 0]]!=""} {
#    puts stderr "Good command: <$command>"
    eval $command
  } else {
    puts stderr "Bad command: $command"
  }
#  addinput -read stdin docommand
  fileevent stdin readable docommand
}

proc bdgraphdefaults {} {
  global col fcol fpat lwid cx cy palette tanchor tfont ufpat

  set col [lindex $palette 0]
  set fcol [lindex $palette 0]
#  set fpat BDfill1
  set ufpat {}
  set lwid 1
  set cx 0
  set cy 0
  set tanchor nw
#  set tfont {-adobe-Times-medium-r-*-*-*-120-*-*-*-*-*-*}
  bdsettextstyle 0 0 1

# TODO Make sure these are sensible defaults
}

proc bdinitialise {width height} {
  global can status

  frame .f -background black 
  pack .f -fill x
  button .f.print -text Postscript -command printcanvas -border 3
  pack .f.print -side left
  text $status -height 1 -width 0 -border 3
  pack $status -fill x -side left -expand 1
  canvas $can -height $height -width $width 
  pack $can 
  bdgraphdefaults
  update
  bind all P printcanvas
  bind all <Any-Button> interaction
  bind all <Any-Key-Return> interaction
}

proc interaction {} {
  global eventhappened

  set eventhappened 1
}

proc printcanvas {} {
  global psfileno 
  
  set file "printout${psfileno}.ps"
  while {[file exists $file]} {
    incr psfileno 1
    set file "printout${psfileno}.ps"
  }
  graphicsprint $file
}

proc graphicsprint {file} {
  global can pspalette status

  $status delete 0.1 end
  $status insert end "Printing to $file ..."
  update
  $can postscript -colormode color -colormap pspalette -file $file
  $status delete 0.1 end
  $status insert end "Finished printing to $file"
  update
}

proc bdsetwinname {name} {
  global w

  wm title $w "$name"
}

proc bdrectangle {left top right bottom} {
  global can col lwid 

  $can create rectangle $left $top $right $bottom -outline $col -width $lwid
}

proc bdarc {x y stang endang rad} {
  
  bdellipse $x $y $stang $endang $rad $rad
}  

proc bdbar {left top right bottom} {
  global can col fcol fpat
  
  $can create rectangle $left $top $right $bottom \
    -outline {} -stipple $fpat -fill $fcol
}

proc bdcircle {x y rad} {
  global can col

  $can create oval \
    [expr $x-$rad] [expr $y-$rad] [expr $x+$rad] [expr $y+$rad] \
    -outline $col
}

proc bdcleardevice {} {
  global can
  $can delete all
  update
}

proc bddrawpoly {npt pts} {
  global can col
  eval [concat $can create line $pts -fill $col]
}

proc bdellipse {x y stang endang xrad yrad} {
  global can col einfo
  
#set einfo for use by bdgetarccoords
  set einfo [list $x $y $stang $endang $xrad $yrad]

  if {($stang==0)&&($endang==360)} {
    $can create oval \
      [expr $x-$xrad] [expr $y-$yrad] [expr $x+$xrad] [expr $y+$yrad] \
      -outline $col
  } else {
    set extent [expr $endang-$stang]
    if {$extent<0} {incr extent 360}
    $can create arc \
      [expr $x-$xrad] [expr $y-$yrad] [expr $x+$xrad] [expr $y+$yrad] \
      -start $stang -extent $extent -style arc -fill $col
  }
}

proc bdfillellipse {x y xrad yrad} {
  global can col fcol fpat
  
  $can create oval \
    [expr $x-$xrad] [expr $y-$yrad] [expr $x+$xrad] [expr $y+$yrad] \
    -outline $col -fill $fcol -stipple $fpat
}

proc bdfillpoly {npt pts} {
  global can col fcol fpat
  
  eval [concat $can create polygon $pts -fill \"$fcol"\ -stipple \"$fpat\"]
  eval [concat $can create line $pts [lrange $pts 0 1] -fill \"$col\"]
}

proc calccoords {x y stang endang xrad yrad} {
  global pi

  set sx [expr round($x+$xrad*cos($stang*$pi/180))]
  set sy [expr round($y-$yrad*sin($stang*$pi/180))]
  set ex [expr round($x+$xrad*cos($endang*$pi/180))]
  set ey [expr round($y-$yrad*sin($endang*$pi/180))]
  list $x $y $sx $sy $ex $ey
}

proc bdgetarccoords {} {
  global einfo

  puts stdout [eval [concat calccoords $einfo]]  
  flush stdout
}

proc bdgetaspectratio {} {

  puts stdout "1 1"
  flush stdout
}

proc bdgetmaxcolor {} {
  global palette

  puts stdout [expr [llength $palette]-1]
  flush stdout
}

proc bdgetmaxx {} {
  global can

  puts stdout [expr [winfo reqwidth $can]-1]
  flush stdout
}

proc bdgetmaxy {} {
  global can

  puts stdout [expr [winfo reqheight $can]-1]
  flush stdout
}

proc bdline {x1 y1 x2 y2} {
  global cx cy can col lwid lpat

  $can create line $x1 $y1 $x2 $y2 -fill $col -width $lwid
#TODO lpat
}
proc bdlinerel {dx dy} {
  global cx cy

  bdlineto [expr $cx+$dx] [expr $cy+$dy]
}

proc bdlineto {x y} {
  global cx cy can col lwid lpat
 
  $can create line $cx $cy $x $y -fill $col -width $lwid
  set cx $x
  set cy $y
#TODO lpat
}
  
proc bdmoverel {dx dy} {
  global cx cy

  bdmoveto [expr $cx+$dx] [expr $cy+$dy]
}

proc bdmoveto {x y} {
  global cx cy
 
  set cx $x
  set cy $y
}

proc bdouttext {text} {
  global can col cx cy tanchor tfont

  set id [$can create text $cx $cy -text "$text" -font "$tfont" \
    -fill $col -anchor $tanchor]
  set box [$can bbox $id]
  set cx [lindex $box 2]
}

proc bdouttextxy {x y text} {
  global can col tanchor tfont

  $can create text $x $y -text "$text" -font $tfont \
    -fill $col -anchor $tanchor
}

proc bdpieslice {x y stang endang rad} {
  bdsector $x $y $stang $endang $rad $rad
}

proc bdrectangle {left top right bottom} {
  global can col lwid lpat

  $can create rectangle $left $top $right $bottom -outline $col -width $lwid
#TODO lpat
}

proc bdsector {x y stang endang xrad yrad} {
  global can col fcol fpat
  
  if {($stang==0)&&($endang==360)} {
    bdfillellipse $x $y $xrad $yrad
  } else {
    set extent [expr $endang-$stang]
    if {$extent<0} {incr extent 360}
    $can create arc \
      [expr $x-$xrad] [expr $y-$yrad] [expr $x+$xrad] [expr $y+$yrad] \
      -start $stang -extent $extent -outline $col -fill $fcol -stipple $fpat
  }
}

proc bdsetbkcolor {newcol} {
  global palette col can

  $can configure -background [lindex $palette $newcol]
}

proc bdsetcolor {newcol} {
  global palette col

  set col [lindex $palette $newcol]
}


proc bdsetfillpattern {newcol pattern} {
  global palette fpat fcol ufpat ufillno

  set fcol [lindex $palette $newcol]
#  blt_bitmap define BDfill$ufillno [list {8 8} $pattern]
#  set ufpat BDfill$ufillno
    set fpat {}
  set fpat $ufpat
}


# From bgiwin.pas
#    EmptyFill           = 0;         { fill patterns }
#    SolidFill           = 1;
#    LineFill            = 2;
#    LtSlashFill         = 3;
#    SlashFill           = 4;
#    BkSlashFill         = 5;
#    LtBkSlashFill       = 6;
#    HatchFill           = 7;
#    XHatchFill          = 8;
#    InterleaveFill      = 9;
#    WideDotFill         = 10;
#    CloseDotFill        = 11;
#    UserFill            = 12;

# blt_bitmap define BDfill0 {{8 8} {00 00 00 00 00 00 00 00}}
# blt_bitmap define BDfill1 {{8 8} {FF FF FF FF FF FF FF FF}}
# blt_bitmap define BDfill2 {{8 8} {FF 00 FF 00 FF 00 FF 00}}
# blt_bitmap define BDfill3 {{8 8} {22 11 88 44 22 11 88 44}}
# blt_bitmap define BDfill4 {{8 8} {cc 66 33 99 cc 66 33 99}}
# blt_bitmap define BDfill5 {{8 8} {33 66 cc 99 33 66 cc 99}}
# blt_bitmap define BDfill6 {{8 8} {88 11 22 44 88 11 22 44}}
# blt_bitmap define BDfill7 {{8 8} {44 ff 44 44 44 ff 44 44}}
# blt_bitmap define BDfill8 {{8 8} {55 ff 55 ff 55 ff 55 ff}}
# blt_bitmap define BDfill9 {{8 8} {55 55 55 55 55 55 55 55}}
# blt_bitmap define BDfill10 {{8 8} {00 00 44 00 00 00 44 00}}
# blt_bitmap define BDfill11 {{8 8} {11 00 44 00 11 00 44 00}}

proc bdsetfillstyle {newpat newcol} {
  global palette fpat fcol ufpat

  set fcol [lindex $palette $newcol]
  if {$newpat>13} {return}
  if {$newpat==12} {
    if {$ufpat!=""} {
      set fpat $ufpat
    } else {
      return
    }
  } else {
#    set fpat BDfill$newpat
      set fpat {}
  }
# TODO set fpat from newpat
}


proc bdsetlinestyle {style pattern thickness} {
  global lpat lwid

  set lwid $thickness
# TODO set lpat from style
}

proc bdsettextjustify {horiz vert} {
  global tanchor
# From bgiwin.pas
#   LeftText            = 0;      { text justification constants }
#    CenterText          = 1;
#    RightText           = 2;
#    BottomText          = 0;
#    TopText             = 2;
  set tanchor [lindex {s {} n} $vert][lindex {w {} e} $horiz]
  if {$tanchor==""} {set tanchor center}
}

proc bdsettextstyle {font direction size} {
  global tdir tfont
  
  set fontname [switch $font {
     0 {format times-medium-r}
     1 {format times-bold-r}
     2 {format times-medium-i}
     3 {format times-bold-o}
     4 {format helvetica-medium-r}
     5 {format helvetica-bold-r}
     6 {format helvetica-medium-o}
     7 {format helvetica-bold-o}
     default {format times-medium-r}
    }]
  
  set tdir $direction

  set fontsize [switch $size {
     0 {format 100}
     1 {format 120}
     2 {format 180}
     3 {format 240}
     default {format 120}
    }]
  set tfont "-adobe-${fontname}-*-*-*-${fontsize}-*-*-*-*-*-*"
}

proc bdwaitinput {} {
  global eventhappened

  set eventhappened 0
  tkwait variable eventhappened
  puts stdout "1 0"
}

  fileevent stdin readable docommand
# addinput -read stdin docommand
